#!/usr/bin/env perl
# created on 2016-08-20

use warnings;
use strict;
use 5.010;
use IO::Handle ();


use Bio::Gonzales::Util::Log;

my $l = Bio::Gonzales::Util::Log->new;

$l->fatal("ERROR1");
$l->fatal_croak("ERROR2");

say STDERR "BERLE";

